# frozen_string_literal: true

require "is/inspectable"
require "pathname"

module Core
  module Watch
    # [public] Contains changes to a watched system, organized by operation.
    #
    class Diff
      include Is::Inspectable

      OPERATIONS = %i[added changed removed].freeze

      def initialize
        @changes = {}
      end

      OPERATIONS.each do |operation|
        # [public]
        #
        class_eval <<~CODE, __FILE__, __LINE__ + 1
          def #{operation}(path)
            @changes[path] = #{operation.inspect}
          end
        CODE
      end

      # [public] Call the given block once with each changed path and operation.
      #
      def each_change(&block)
        return to_enum(:each_change) unless block

        @changes.each_pair(&block)
      end

      # [public] Call the given block once with each changed path.
      #
      def each_changed_path(&block)
        return to_enum(:each_changed_path) unless block

        @changes.each_key(&block)
      end

      # [public] Return `true` if the diff contains changes.
      #
      def changed?
        @changes.any?
      end

      # [public] Return `true` if `path` is included in the changes.
      #
      def include?(path)
        @changes.include?(Pathname(path))
      end

      # [public] Return the operation for the given path.
      #
      def operation(path)
        @changes[path]
      end
    end
  end
end
