# frozen_string_literal: true

require "is/inspectable"
require "pathname"

require "core/async"

require_relative "callback"
require_relative "snapshot"
require_relative "status"

module Core
  module Watch
    # [public] The system watcher.
    #
    class System
      include Is::Async
      include Is::Inspectable
      inspects without: [:@mutex, :@working]

      # [public] The configured polling interval.
      #
      attr_accessor :interval

      # [public] The status object.
      #
      attr_reader :status

      # [public] The current latency.
      #
      attr_reader :latency

      def initialize(interval: 0.1, strategy: :digest)
        @interval = interval
        @status = Status.new
        @mutex = Mutex.new
        @watched = []
        @ignored = []
        @callbacks = []
        @working = false
        @latency = nil

        @strategy = case strategy.to_sym
        when :digest
          require_relative "strategies/digest"
          Strategies::Digest
        when :timestamp
          require_relative "strategies/timestamp"
          Strategies::Timestamp
        else
          raise ArgumentError, "unknown watch strategy `#{strategy.inspect}'"
        end
      end

      # [public] Watch `path` for changes.
      #
      def watch(path)
        path = Pathname(path).expand_path

        @mutex.synchronize do
          unless @watched.include?(path)
            mutate do
              @watched << path
            end
          end
        end
      end

      # [public] Ignore changes at `path`.
      #
      def ignore(path)
        path = path.to_s

        @mutex.synchronize do
          unless @ignored.include?(path)
            mutate do
              @ignored << path
            end
          end
        end
      end

      # [public] Register a callback to be called when a change is detected. The callback will only be called once per
      # tick and will be given a `Core::Watch::Diff` instance. If `matcher` is passed, the callback will only be called
      # for matching changes. The matcher can be a string or any object that responds to `match?`.
      #
      def callback(matcher = nil, &block)
        raise ArgumentError, "expected a block" if block.nil?

        @mutex.synchronize do
          @callbacks << Callback.build(matcher, &block)
        end
      end

      # [public] Start the watcher.
      #
      def start
        await do
          unless @status.running?
            @future = run
          end
        end
      end

      # [public] Stop the watcher.
      #
      def stop(force: false)
        if !@working || force
          @future&.cancel
        end

        @status.stopped!
      end

      # [public] Pause the watcher, causing it to continue running but not invoke callbacks.
      #
      def pause
        @status.paused!
      end

      # [public] Resume a paused watcher.
      #
      def resume
        @status.running!
      end

      private def run
        async {
          @status.running!

          rebuild_snapshot

          until @status.stopped?
            sleep(@interval)

            started = current_time

            # Don't process this tick if the watch was paused or stopped while sleeping.
            #
            next if @status.paused? || @status.stopped?

            # Do the work, updating the snapshot in the process.
            #
            invoke_callbacks(@snapshot.diff)

            # Measure the latency for this tick.
            #
            @latency = [0, current_time - started - @interval].max
          end
        }
      end

      private def mutate
        yield

        if @status.running?
          rebuild_snapshot
        end
      end

      private def rebuild_snapshot
        @snapshot = Snapshot.new(*@watched, ignore: @ignored, strategy: @strategy)
      end

      private def invoke_callbacks(diff)
        return unless diff.changed?

        await do
          @working = true
          @callbacks.each do |callback|
            maybe_invoke_callback(callback, diff)
          end
        end
      ensure
        @working = false
      end

      private def maybe_invoke_callback(callback, diff)
        diff.each_change do |path, event|
          if callback.match?(path)
            return async { callback.call(diff) }
          end
        end
      end

      private def current_time
        Process.clock_gettime(Process::CLOCK_MONOTONIC)
      end
    end
  end
end
